"""
    API class for working with API (HTTP).
    Sends HTTP requests, handles API methods.
"""
import requests

from typing import Optional

from gatey_sdk.auth import Auth
from gatey_sdk.response import Response
from gatey_sdk.exceptions import GateyApiError, GateyApiAuthError
from gatey_sdk.utils import remove_trailing_slash
from gatey_sdk.consts import (
    API_DEFAULT_SERVER_PROVIDER_URL,
    API_DEFAULT_SERVER_EXPECTED_VERSION,
)


class Api:
    """
    Wrapper for API methods, HTTP sender.
    """

    # URL of the API.
    # Can be changed for Self-Hosted servers.
    _api_server_provider_url = API_DEFAULT_SERVER_PROVIDER_URL

    # Timeout for requests.
    _api_server_requests_timeout = 7

    # Version that expected from the API.
    _api_server_expected_version = API_DEFAULT_SERVER_EXPECTED_VERSION

    # `Auth` instance that provides authentication fields.
    _auth_provider: Auth = None

    def __init__(self, auth: Optional[Auth] = None):
        """
        :param auth: Auth provider as the `Auth` instance.
        """
        if auth and not isinstance(auth, Auth):
            raise TypeError(
                "Auth must be an instance of `Auth`! You may not pass auth as it will be initialise blank internally in `Api`."
            )
        self._auth_provider = auth if auth else Auth()

    def method(
        self,
        name: str,
        *,
        send_access_token: bool = False,
        send_project_auth: bool = False,
        **kwargs,
    ) -> Response:
        """
        Executes API method with given name.
        And then return response from it.
        :param name: Name of the method to call.
        """

        http_params = kwargs.copy()
        if send_access_token and self._auth_provider:
            if self._auth_provider.access_token:
                http_params.update({"access_token": self._auth_provider.access_token})

        if send_project_auth and not send_access_token and self._auth_provider:
            if self._auth_provider.project_id:
                http_params.update({"project_id": self._auth_provider.project_id})
            if self._auth_provider.server_secret:
                http_params.update({"server_secret": self._auth_provider.server_secret})
            if (
                self._auth_provider.client_secret
                and not self._auth_provider.server_secret
            ):
                http_params.update({"client_secret": self._auth_provider.client_secret})

        # Build URL where API method is located.
        api_server_method_url = f"{self._api_server_provider_url}/{name}"

        # Send HTTP request.
        http_response = requests.get(
            url=api_server_method_url,
            params=http_params,
            timeout=self._api_server_requests_timeout,
        )

        # Wrap HTTP response in to own Response object.
        response = Response(http_response=http_response)

        # Raise exception if there is any error returned with Api.
        self._process_error_and_raise(method_name=name, response=response)

        return response

    def change_api_server_provider_url(self, provider_url: str) -> None:
        """
        Updates API server provider URL.
        Used for self-hosted servers.
        :param provider_url: URL of the server API provider.
        """
        provider_url = remove_trailing_slash(provider_url)
        self._api_server_provider_url = provider_url

    def change_api_server_timeout(self, timeout: int) -> None:
        """
        Updates API timeout for requests.
        :param timeout: New timeout
        """
        self._api_server_requests_timeout = timeout

    def change_api_server_expected_version(self, version: str) -> None:
        """
        Updates API version.
        :param version: Version of API.
        """
        self._api_server_expected_version = version

    def _process_error_and_raise(self, method_name: str, response: Response) -> None:
        """
        Processes error, and if there is any error, raise ApiError exception.
        """
        error = response.raw_json().get("error")
        if error:
            # If there is an error.

            # Query error fields.
            error_message = error.get("message")
            error_code = error.get("code")
            error_status = error.get("status")

            # If invalid request by validation error, there will be additional error information in "exc" field of the error.
            if error_code == 3 and "exc" in error:
                error_message = f"{error_message} Additional exception information: {error.get('exc')}"

            # Raise ApiError exception.
            message = f"Failed to call API method {method_name}! Error code: {error_code}. Error message: {error_message}"
            raise GateyApiError(
                message=message,
                error_code=error_code,
                error_message=error_message,
                error_status=error_status,
                response=response,
            )

    def do_auth_check(self) -> bool:
        """
        Checks authentication with API.
        Returns is it successfully or no.
        """
        try:
            self.do_hard_auth_check()
        except (GateyApiError, GateyApiAuthError):
            return False
        return True

    def do_hard_auth_check(self) -> bool:
        """
        Checks authentication with API.
        Raises API error exception if unable to authenticate you.
        """
        try:
            self.method(
                "project.checkAuthority",
                send_access_token=False,
                send_project_auth=True,
            )
        except GateyApiError as api_error:
            if api_error.error_code == 7:
                raise GateyApiAuthError(
                    "You are entered incorrect project secret (client or server)! Please review your SDK settings! (See previous exception to see more described information)"
                )
            if api_error.error_code == 8:
                raise GateyApiAuthError(
                    "You are entered not existing project id! Please review your SDK settings! (See previous exception to see more described information)"
                )
            raise GateyApiAuthError(
                "There is unknown error while trying to check auth (do_auth)! (See previous exception to see more described information)"
            )



"""
    Utils methods.
"""

import sys
import platform
import tokenize

from typing import Any, Dict, List, Callable, Optional, Union
from types import TracebackType

from gatey_sdk.consts import SDK_INFORMATION_DICT
from gatey_sdk.exceptions import (
    GateyApiError,
    GateyTransportError,
    GateyTransportImproperlyConfiguredError,
)
from gatey_sdk.consts import (
    EXC_ATTR_SHOULD_SKIP_SYSTEM_HOOK,
    EXC_ATTR_WAS_HANDLED,
    RUNTIME_NAME,
)


def wrap_in_exception_handler(
    *,
    reraise: bool = True,
    exception: Optional[BaseException] = None,
    ignored_exceptions: Optional[List[BaseException]] = None,
    on_catch_exception: Optional[Callable] = None,
    skip_global_handler_on_ignore: bool = False,
) -> Callable:
    """
    Decorator that catches the exception and captures it as Gatey exception.
    :param reraise: If False, will not raise the exception again, application will not fall (WARNING: USE THIS WISELY TO NOT GET UNEXPECTED BEHAVIOR)
    :param exception: Target exception type to capture.
    :param ignored_exceptions: List of exceptions that should not be captured.
    :param on_catch_exception: Function that will be called when an exception is caught.
    :param skip_global_handler_on_ignore: If true, will skip global exception handler if exception was ignored.
    """

    # Default target exception value.
    if exception is None:
        exception = BaseException

    # Default value for ignored exception list (Do not ignore any exceptions)
    if ignored_exceptions is None:
        ignored_exceptions = []

    def decorator(function: Callable):
        def wrapper(*args, **kwargs):
            # Gets called when `decorated` function get called.
            try:
                # This will simply return function result if there is no exception.
                return function(*args, **kwargs)
            except exception as e:
                # There is any exception that we should handle occurred.

                # Typed.
                e: BaseException = e

                # Do not handle ignored exceptions.
                if exception_is_ignored(e, ignored_exceptions):
                    if skip_global_handler_on_ignore:
                        # If we should skip global exception handler.
                        setattr(e, EXC_ATTR_SHOULD_SKIP_SYSTEM_HOOK, True)
                    raise e

                # Call catch event.
                if callable(on_catch_exception):
                    on_catch_exception(e)
                    # Mark as handled.
                    setattr(e, EXC_ATTR_WAS_HANDLED, True)

                # Raise exception again if we expected that.
                if reraise is True:
                    raise e

        return wrapper

    return decorator


def register_system_exception_hook(
    hook: Callable, skip_internal_exceptions: bool = True
):
    """
    Register exception hook for system.
    :param hook: Will be called when exception triggered.
    """

    def _system_exception_hook_handler(
        exception_type: type[BaseException],
        exception: BaseException,
        traceback: TracebackType,
    ):
        # System exception hook handler.

        was_handled = hasattr(exception, EXC_ATTR_WAS_HANDLED)
        if not was_handled and not hasattr(exception, EXC_ATTR_SHOULD_SKIP_SYSTEM_HOOK):
            # If marked as skipped for system hook.
            try:
                # Try to handle this exception with hook.
                hook(exception=exception)
            except (
                GateyApiError,
                GateyTransportError,
                GateyTransportImproperlyConfiguredError,
            ) as e:
                # If there is any error while processing global exception handler.
                if not skip_internal_exceptions:
                    raise e

        # Default system hook.
        sys.__excepthook__(exception_type, exception, traceback)

    # Register system exception hook.
    sys.excepthook = _system_exception_hook_handler


def exception_is_ignored(
    exception: BaseException, ignored_exceptions: List[BaseException]
) -> bool:
    """
    Returns True if exception should be ignored based on `ignored_exceptions` list.
    """
    exception_type = type(exception)
    for ignored_exception_type in ignored_exceptions:
        if exception_type == ignored_exception_type:
            return True
    return False


def event_dict_from_exception(
    exception: BaseException, skip_vars: bool = True, include_code_context: bool = True
) -> Dict:
    """
    Returns event dictionary of the event (field) from the raw exception.
    Fetches all required information about system, exception.
    """

    # Get raw exception traceback information.
    exception_traceback = getattr(exception, "__traceback__", None)

    # Query traceback information.
    traceback_vars = get_variables_from_traceback(
        traceback=exception_traceback, _always_skip=skip_vars
    )
    traceback_trace = get_trace_from_traceback(
        exception_traceback, include_code_context=include_code_context
    )

    # Get exception type ("BaseException", "ValueError").
    exception_type = get_exception_type_name(exception)
    exception_description = str(exception)
    event_dict = {
        "class": exception_type,
        "description": exception_description,
        "vars": traceback_vars,  # Will be migrated to the traceback context later.
        "traceback": traceback_trace,
    }
    return event_dict


def get_exception_type_name(exception: BaseException) -> str:
    """
    Returns exception type ("BaseException", "ValueError").
    """
    return getattr(type(exception), "__name__", "NoneException")


def get_additional_event_tags(
    include_platform_info: bool = True,
    include_runtime_info: bool = True,
    include_sdk_info: bool = True,
) -> Dict[str, Any]:
    """
    Returns additional event dictionary for tags with event information such as SDK information, platform information etc.
    """
    additional_event_tags = dict()
    if include_sdk_info:
        additional_event_tags.update(SDK_INFORMATION_DICT)
    if include_platform_info:
        additional_event_tags.update(get_platform_event_tags())
    if include_runtime_info:
        additional_event_tags.update(get_runtime_event_tags())
    return additional_event_tags


def remove_trailing_slash(url: str) -> str:
    """
    Removes trailing slash from a URL.
    Example: `http://example.com/` will become `http://example.com` (No trailing slash)

    :param url: The URL to remove trailing slash.
    """
    if not isinstance(url, str):
        raise TypeError("URL must be a string!")

    if url.endswith("/"):
        url = url[:-1]

    return url


def get_trace_from_traceback(
    traceback: TracebackType,
    include_code_context: bool = True,
    code_context_lines_count: int = 5,
    code_context_only_for_tail: bool = True,
) -> List[Dict]:
    """
    Returns trace from the given traceback.
    """

    trace = []

    while traceback is not None:
        # Iterating over traceback with `tb_next`
        frame = traceback.tb_frame
        frame_code = getattr(frame, "f_code", None)
        filename, function = None, None
        if frame_code:
            filename = frame.f_code.co_filename
            function = frame.f_code.co_name

        line_number = traceback.tb_lineno
        trace_element = {
            "filename": filename,
            "name": function or "<unknown>",
            "line": line_number,
            "module": frame.f_globals.get("__name__", None),
        }

        if include_code_context and not code_context_only_for_tail:
            trace_element |= {
                "context": get_context_lines_from_source_code(
                    filename=filename,
                    line_number=line_number,
                    context_lines_count=code_context_lines_count,
                )
            }

        trace.append(trace_element)
        traceback = traceback.tb_next

    if include_code_context and code_context_only_for_tail:
        tail_trace = trace[-1]
        tail_trace["context"] = get_context_lines_from_source_code(
            filename=tail_trace["filename"],
            line_number=tail_trace["line"],
            context_lines_count=5,
        )

    return trace


def get_context_lines_from_source_code(
    filename: str, line_number: int, context_lines_count: int = 5
) -> Dict[str, Union[str, None, List[str]]]:
    """
    Returns context lines from source code file.
    """
    source_code_lines = get_lines_from_source_code(filename=filename)
    bounds_start = max(0, line_number - context_lines_count - 1)
    bounds_end = min(line_number + 1 + context_lines_count, len(source_code_lines))

    strip_line = lambda line: line.strip("\r\n").replace("    ", "\t")
    context_pre, context_target, context_post = [], None, []
    try:
        context_pre = [
            strip_line(line)
            for line in source_code_lines[bounds_start : line_number - 1]
        ]
        context_target = strip_line(source_code_lines[line_number - 1])
        context_post = [
            strip_line(line) for line in source_code_lines[line_number:bounds_end]
        ]
    except IndexError:
        # File was changed?
        pass
    return {"pre": context_pre, "target": context_target, "post": context_post}


def get_lines_from_source_code(filename: str) -> List[str]:
    """
    Returns lines of the code from the source code filename.
    """
    try:
        with tokenize.open(filename=filename) as source_file:
            return source_file.readlines()
    except (OSError, IOError):
        return []


def get_platform_event_tags() -> Dict[str, Any]:
    """
    Returns platform information for event data tags.
    """

    platform_os = platform.system()
    platform_network_name = platform.node()
    platform_event_data_tags = {
        "os": platform_os,
        "os.node": platform_network_name,
        "os.version": platform.version(),  # For major there is `platform.release()`
        "os.bits": platform.architecture()[0],
    }

    return platform_event_data_tags


def get_runtime_event_tags() -> Dict:
    """
    Returns runtime information event data tags.
    """
    runtime_name = RUNTIME_NAME
    runtime_version = sys.version_info
    runtime_version = f"{runtime_version[0]}.{runtime_version[1]}.{runtime_version[2]}-{runtime_version[3]}-{runtime_version[4]}"
    return {
        "runtime.name": runtime_name,
        "runtime.ver": runtime_version,
        "runtime.impl": platform.python_implementation(),
    }


def get_variables_from_traceback(
    traceback: TracebackType, *, _always_skip: bool = False
) -> Dict:
    """
    Returns local and global variables from the given traceback.
    """

    traceback_variables_locals = {}
    traceback_variables_globals = {}

    if traceback and not _always_skip:
        traceback_variables_locals = traceback.tb_frame.f_locals
        traceback_variables_globals = traceback.tb_frame.f_globals

    # Stringify variable values.
    traceback_variables_locals = {
        key: str(value) for key, value in traceback_variables_locals.items()
    }
    traceback_variables_globals = {
        key: str(value) for key, value in traceback_variables_globals.items()
    }

    return {
        "locals": traceback_variables_locals,
        "globals": traceback_variables_globals,
    }



"""
    NOT REFACTORED.
    NOT USED.
"""

from typing import Optional

# from urllib.parse import urlparse
# from urllib.parse import parse_qs


class Auth:
    """
    Wrapper for authentication data (access token, project information for capturing (project id, client / server secret))
    """

    # Access token is used for user authorized calls.
    # Like editing project, or interacting with administration tools.
    access_token: Optional[str] = None

    # Project information for capturing events.
    # Secrets is used to verify calls to the project event capturer.
    project_id: Optional[int] = None
    server_secret: Optional[str] = None
    client_secret: Optional[str] = None

    def __init__(
        self,
        access_token: Optional[str] = None,
        project_id: Optional[int] = None,
        server_secret: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        """
        :param access_token: Access token of the your account with `gatey` scope.
        :param project_id: Project id to capture event for.
        :param server_secret: Secret of the project.
        :param client_secret: Secret of the project.
        """
        self.access_token = access_token
        self.project_id = project_id
        self.server_secret = server_secret
        self.client_secret = client_secret

    def request_oauth_from_stdin(self) -> None:
        """
        Get access token from stdin (IO, user).
        """

        print(
            "\tOpen page in browser, signin and copy url here:",
            self.get_manual_oauth_user_login_url(),
            sep="\n\t\t",
        )

        oauth_redirect_uri = input("\tRedirect URI: ")
        oauth_access_token = self.parse_access_token_from_redirect_uri(
            redirect_uri=oauth_redirect_uri
        )

        print(f"\tSuccessfully grabbed access token: {oauth_access_token}")
        self.access_token = oauth_access_token

    @staticmethod
    def get_manual_oauth_user_login_url(
        client_id: int = 1,
        scope: str = "gatey",
        response_type: str = "token",
        redirect_uri: str = "https://florgon.space/oauth/blank",
        oauth_screen_url: str = "https://florgon.space/oauth/authorize",
    ) -> str:
        """
        Returns url for OAuth user login manually.
        """
        return f"{oauth_screen_url}?client_id={client_id}&redirect_uri={redirect_uri}&scope={scope}&response_type={response_type}"

    @staticmethod
    def parse_access_token_from_redirect_uri(redirect_uri: str) -> str | None:
        """
        Returns token from redirect uri (OAuth) or None if not found there.
        """
        url = redirect_uri.split("#token=")
        return url[1] if len(url) > 1 else None



"""
    Custom exceptions that may occur while working with SDK.
"""

from gatey_sdk.response import Response
from gatey_sdk.consts import EXC_ATTR_IS_INTERNAL


class GateyError(Exception):
    """
    Super class for Gatey exceptions.
    """


class GateyApiError(GateyError):
    """
    Raised when there is any error with response.
    Means API return error (not success).
    """

    def __init__(
        self,
        message: str,
        error_code: int,
        error_message: str,
        error_status: int,
        response: Response,
    ):
        """
        :param message: Message of the exception.
        :param error_code: API error code
        :param error_message: API error message.
        :param error_status: API error status (HTTP status, from the API `status` error field).
        :param response: API response.
        """
        super().__init__(message)
        self.error_code = error_code
        self.error_message = error_message
        self.error_status = error_status
        self.response = response
        setattr(self, EXC_ATTR_IS_INTERNAL, True)


class GateyApiAuthError(GateyError):
    """
    Raised when there is any error in the auth to the API.
    """

    def __init__(self, message: str):
        """
        :param message: Message of the exception.
        """
        super().__init__(message)
        setattr(self, EXC_ATTR_IS_INTERNAL, True)


class GateyTransportError(GateyError):
    """
    Raised when there is any error in the transport.
    For example, raised when `FuncTransport` function raises any exceptions.
    """

    def __init__(self, message: str):
        """
        :param message: Message of the exception.
        """
        super().__init__(message)
        setattr(self, EXC_ATTR_IS_INTERNAL, True)


class GateyTransportImproperlyConfiguredError(GateyTransportError):
    """
    Raised when there is any error in the transport configuration.
    For example, raised when no project id or client / server secret.
    """

    def __init__(self, message: str):
        """
        :param message: Message of the exception.
        """
        super().__init__(message)
        setattr(self, EXC_ATTR_IS_INTERNAL, True)
